#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Package bvpy documentation build configuration file, created by
# sphinx-quickstart on Tue Jul  9 22:26:36 2013.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import os
import re
import sys
import shutil
from sphinx.application import Sphinx

sys.path.insert(0, os.path.abspath('.')) # add doc/ to the path
#sys.path.insert(0, os.path.abspath('./tutorials'))  # Add the `tutorials` directory

from utils.convert_notebook import process_notebook, clean_and_reset_notebook

# ---------------------- Process Tutorials ---------------------- #
source_tutorials_dir = '../tutorials'
tutorial_output_dir = './tutorials'
tutorial_rst_path = "tutorials.rst"

# List of tutorial notebooks to process
notebooks = [
    'bvpy_tutorial_1_hello_world.ipynb',
    'bvpy_tutorial_2_domains.ipynb',
    'bvpy_tutorial_3_vforms.ipynb',
    'bvpy_tutorial_4_bnd_cond.ipynb',
    'bvpy_tutorial_5_reaction_diffusion.ipynb',
    'bvpy_tutorial_6_linear_elasticity.ipynb',
    'bvpy_tutorial_7_hyper_elasticity.ipynb',
    'bvpy_tutorial_8_pyvista.ipynb',
    'bvpy_tutorial_9_gmsh.ipynb',
    'bvpy_tutorial_10_morphoelasticity.ipynb',
]

# Ensure the tutorials directory exists
os.makedirs(tutorial_output_dir, exist_ok=True)

# Copy supporting folders (images, tutorial_data)
folders_to_copy = ['images', 'tutorial_data']
for folder_name in folders_to_copy:
    source_folder = os.path.join(source_tutorials_dir, folder_name)
    target_folder = os.path.join(tutorial_output_dir, folder_name)
    if os.path.exists(source_folder):
        shutil.rmtree(target_folder, ignore_errors=True)  # Clear old data
        shutil.copytree(source_folder, target_folder)
        print(f"Copied {source_folder} -> {target_folder}")
    else:
        print(f"Source folder {source_folder} not found. Skipping...")

# Remove the old tutorials.rst file
if os.path.exists(tutorial_rst_path):
    print(f"Removing old {tutorial_rst_path}")
    os.remove(tutorial_rst_path)

# Generate tutorials.rst with a toctree
tutorial_rst_lines = [
    ".. _tutorials:\n\n",
    "Tutorials\n*********\n\n",
    "Download the tutorial notebooks and accompanying data as a ZIP file: "
    "`tutorials.zip <https://mosaic.gitlabpages.inria.fr/bvpy/tutorials.zip>`_\n\n\n",
    ".. toctree::\n",
    "    :titlesonly:\n",
    "    :maxdepth: 1\n\n",
]

# Process each notebook
for notebook in notebooks:
    input_path = os.path.join(source_tutorials_dir, notebook)
    output_path = os.path.join(tutorial_output_dir, notebook)
    process_notebook(input_notebook_path=input_path, output_notebook_path=output_path, dynamic_plot=False)
    tutorial_rst_lines.append(f"    tutorials/{notebook.replace('.ipynb', '')}\n")

# Write the tutorials.rst file
with open(tutorial_rst_path, "w", encoding="utf-8") as f:
    f.writelines(tutorial_rst_lines)
print(f"Wrote {tutorial_rst_path}")

# -------------------- Clean Tutorials After Build -------------------- #
def clean_doc(app: Sphinx, exception):
    """
    Remove temporary folders created during the build process.
    Cleans up the tutorials folder and the _dvlpt folder after the build is complete.
    """
    # clean the tutorials/ folder in order to zip its content in CI/CD
    if os.path.exists(tutorial_output_dir):
        print(f"Cleaning up notebooks in {tutorial_output_dir}")
        for notebook in notebooks:
            notebook_path = os.path.join(tutorial_output_dir, notebook)
            clean_and_reset_notebook(input_notebook_path=notebook_path, output_notebook_path=notebook_path)

    # Clean up the _dvlpt folder
    dev_folder = os.path.abspath(os.path.join(project_root, "doc", "_dvlpt"))
    if os.path.exists(dev_folder):
        print(f"Cleaning up {dev_folder}")
        shutil.rmtree(dev_folder, ignore_errors=True)

# ---------------------- Sphinx Setup ---------------------- #
def setup(app):
    """Hook custom setup functionality into Sphinx."""
    app.add_css_file('css/custom.css')  # Add custom CSS
    app.connect('build-finished', clean_doc)

# ------------------- General Configuration ------------------- #
project_root = os.path.dirname(os.getcwd())
sys.path.insert(0, os.path.join(project_root, 'src'))  # Add src/ to the path

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.doctest',
    'sphinx.ext.graphviz',
    'sphinx.ext.ifconfig',
    'sphinx.ext.inheritance_diagram',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.napoleon',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    'sphinx_tabs.tabs',
    'nbsphinx',
    'recommonmark',
    'pyvista.ext.plot_directive',
    'pyvista.ext.viewer_directive',
    'sphinx_design',
]

# Enable plotly figures and add custom CSS for hiding cells with `remove_cell` tags
nbsphinx_prolog = r"""
.. raw:: html

    <script src='http://cdnjs.cloudflare.com/ajax/libs/require.js/2.1.10/require.min.js'></script>
    <script>require=requirejs;</script>
"""

nbsphinx_timeout = -1
todo_include_todos = True
autosummary_generate = True
intersphinx_mapping = {
    'python': ('https://docs.python.org/3.4', None),
    'dolfin': ('https://fenicsproject.org/olddocs/dolfin/2019.1.0/python/', None),
    'ufl': ('https://fenics.readthedocs.io/projects/ufl/en/latest/', None),
}
inheritance_node_attrs = dict(shape='ellipse', fontsize=12, color='orange', style='filled')

# Templates, source files, and static paths
templates_path = ['_templates']
source_suffix = '.rst'
master_doc = 'index'
exclude_patterns = ['build', 'dist', '**.ipynb_checkpoints']

# General information
pyproject_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", "pyproject.toml")
# Open and search for the version in pyproject.toml
with open(pyproject_path, encoding="utf-8") as f:
    content = f.read()
    match = re.search(r'^version\s*=\s*["\'](.*?)["\']', content, re.MULTILINE)
    version = match.group(1) if match else "undefined"
    release = version  # Use this for detailed versioning

project = "bvpy"
copyright = "2020, bvpy"

# Syntax highlighting style
pygments_style = 'sphinx'

# ------------------- HTML Output ------------------- #
html_theme = "sphinx_rtd_theme"
html_logo = "_static/logo_side.png"
html_favicon = "_static/favicon.ico"
html_static_path = ['_static']
htmlhelp_basename = "bvpydoc"

# ------------------- LaTeX Output ------------------- #
latex_elements = {
    'preamble': r'''
        \usepackage{amsmath}
        \usepackage{amssymb}
    ''',
}
# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
latex_documents = [
    ('index', 'bvpy', 'bvpy Documentation', ['moi'], 1),
]

# ------------------- Texinfo Output ------------------- #
# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    ('index', 'bvpy', 'bvpy Documentation', 'moi', 'bvpy', 'A package implementing Boundary Value Problem.',
     'Miscellaneous'),
]

# ------------------- Apidoc generation ------------------- #
try:
    from sphinx.ext.apidoc import main
except ImportError:
    from sphinx.apidoc import main


destdir = os.path.abspath(os.path.join(project_root, "doc", "_dvlpt"))

if not os.path.isdir(destdir):
    os.makedirs(destdir)

src_dir = os.path.abspath(os.path.join(project_root, "src"))
main(['-e', '-o', destdir, '-d', '4', '-s', source_suffix[1:], '--force', src_dir])
